// Panel3D.java

import java.awt.*;

import java.io.*;

import java.text.*;

import java.util.*;

import javax.swing.*;

public class Panel3D extends JPanel
{
   // paintComponent() color constants

   private final static Color PANEL_COLOR = new Color (255, 255, 225);
   private final static Color XAXIS_COLOR = new Color (192, 0, 0);
   private final static Color YAXIS_COLOR = new Color (0, 192, 0);
   private final static Color ZAXIS_COLOR = new Color (0, 0, 192);

   // clip() and code() category constants

   private final static int LEFT = 1;
   private final static int RIGHT = 2;
   private final static int TOP = 4;
   private final static int BOTTOM = 8;

   // current transformation matrix

   private Matrix transform;

   // loaded vertices from model file

   private HashMap vertices;

   // loaded edges from model file

   private Vector edges;

   // current drawing position

   private double cx, cy, cz;

   // viewport height -- established in setViewport(); used in showLine() to 
   // invert the y coordinate so that (0,0) represents the lower-left corner
   // instead of the upper-left corner

   private int height;

   // viewport center coordinates and viewport (x,y) sizes (divided by 2) --
   // established in setViewPort(); used in showLine()

   private double vcx, vcy, vsx, vsy;

   public Panel3D (int width)
   {
      // Create viewing transformation matrix, initialized to the identity
      // matrix.

      transform = new Matrix (4, 4);
      identity ();

      vertices = new HashMap ();

      edges = new Vector ();

      setViewport (0, 0, width-1, width-1);

      setPreferredSize (new Dimension (width, width));
      setMinimumSize (getPreferredSize ());

      setBorder (BorderFactory.createEtchedBorder ());
   }

   public void load (String filename) throws IOException, ParseException
   {
      StreamTokenizer st;
      st = new StreamTokenizer (new FileReader (filename));

      // Recognize C++ comments.

      st.slashSlashComments (true);

      // End-of-lines are to be returned as tokens.

      st.eolIsSignificant (true);

      // Ignore leading blank lines;

      while (st.nextToken () == StreamTokenizer.TT_EOL);

      // Exit on premature end of file.

      if (st.ttype == StreamTokenizer.TT_EOF)
          return;

      // Reset vertices and edges.

      vertices.clear ();
      edges.clear ();

      // Parse vertices.

      do
      {
         if (st.ttype != StreamTokenizer.TT_WORD)
             throw new ParseException ("Vertex name expected", -1);

         String name = st.sval;

         if (st.nextToken () != StreamTokenizer.TT_NUMBER)
             throw new ParseException ("Vertex x value expected", -1);

         double x = st.nval;

         if (st.nextToken () != StreamTokenizer.TT_NUMBER)
             throw new ParseException ("Vertex y value expected", -1);

         double y = st.nval;

         if (st.nextToken () != StreamTokenizer.TT_NUMBER)
             throw new ParseException ("Vertex z value expected", -1);

         double z = st.nval;

         vertices.put (name, new Vertex (x, y, z));

         if (st.nextToken () != StreamTokenizer.TT_EOL)
             throw new ParseException ("End of line expected", -1);

         if (st.nextToken () == StreamTokenizer.TT_EOL)
             break;
      }
      while (true);

      // Ignore intervening blank lines;

      while (st.nextToken () == StreamTokenizer.TT_EOL);

      // Parse edges.

      do
      {
         if (st.ttype != StreamTokenizer.TT_WORD)
             throw new ParseException ("Start vertex name expected", -1);

         String sname = st.sval;

         if (st.nextToken () != StreamTokenizer.TT_WORD)
             throw new ParseException ("End vertex name expected", -1);

         String ename = st.sval;

         edges.add (new Edge (sname, ename));

         if (st.nextToken () != StreamTokenizer.TT_EOL)
             throw new ParseException ("End of line expected", -1);

         if (st.nextToken () == StreamTokenizer.TT_EOL ||
             st.ttype == StreamTokenizer.TT_EOF)
             break;
      }
      while (true);

      repaint ();
   }

   public void lookFrom (double x, double y, double z)
   {
      identity ();

      translate (-x, -y, -z);

      rotateX (90.0);

      double angle;
      if (x >= 0 && y >= 0)
          angle = Math.toDegrees (Math.atan (x/y)) + 180.0;
      else
      if (x < 0 && y >= 0)
          angle = Math.toDegrees (Math.atan (x/y)) - 180.0;
      else
      if (x >= 0 && y < 0)
          angle = Math.toDegrees (Math.atan (x/y));
      else
          angle = Math.toDegrees (Math.atan (x/y));

      rotateY (-angle);

      angle = Math.toDegrees (Math.atan (z/Math.sqrt (x*x+y*y)));
      rotateX (-angle);

      scale (1.0, 1.0, -1.0);
   }

   public void paintComponent (Graphics g)
   {
      super.paintComponent (g);

      // Paint panel background.

      g.setColor (PANEL_COLOR);
      g.fillRect (0, 0, getWidth (), getHeight ());

      // Draw x axis.

      g.setColor (XAXIS_COLOR);
      moveto (-2.0, 0.0, 0.0);
      lineto (g, 2.0, 0.0, 0.0);

      // Draw dash to identify negative x axis.

      g.setColor (Color.black);
      moveto (-2.1, 0.0, 0.0);
      lineto (g, -2.2, 0.0, 0.0);

      // Draw y axis.

      g.setColor (YAXIS_COLOR);
      moveto (0.0, -2.0, 0.0);
      lineto (g, 0.0, 2.0, 0.0);

      // Draw dash to identify negative y axis.

      g.setColor (Color.black);
      moveto (0.0, -2.1, 0.0);
      lineto (g, 0.0, -2.2, 0.0);

      // Draw z axis.

      g.setColor (ZAXIS_COLOR);
      moveto (0.0, 0.0, -2.0);
      lineto (g, 0.0, 0.0, 2.0);

      // Draw dash to identify negative z axis.

      g.setColor (Color.black);
      moveto (0.0, 0.0, -2.1);
      lineto (g, 0.0, 0.0, -2.2);

      // Draw model.

      g.setColor (Color.black);

      Enumeration e = edges.elements ();

      while (e.hasMoreElements ())
      {
         Edge edge = (Edge) e.nextElement ();

         Vertex v1 = (Vertex) vertices.get (edge.startVertex);
         moveto (v1.x, v1.y, v1.z);

         Vertex v2 = (Vertex) vertices.get (edge.endVertex);
         lineto (g, v2.x, v2.y, v2.z);
      }
   }

   public void perspective (double ds)
   {
      Matrix p = new Matrix (4, 4);
      p.setValue (0, 0, ds);
      p.setValue (1, 1, ds);
      p.setValue (2, 2, 1.0);
      p.setValue (3, 3, 1.0);

      transform = transform.multiply (p);

      repaint ();
   }

   private void clip (Graphics g, double x1, double y1, double z1, double x2,
                      double y2, double z2)
   {
      int c, c1, c2;
      double t, x = 0.0, y = 0.0, z = 0.0;

      c1 = code (x1, y1, z1);
      c2 = code (x2, y2, z2);

      while (c1 != 0 || c2 != 0)
      {
         if ((c1 & c2) != 0)
             return;

         c = c1;
         if (c == 0)
             c = c2;

         if ((c & LEFT) != 0)
         {
             t = (z1 + x1) / ((x1 - x2) - (z2 - z1));
             z = t * (z2 - z1) + z1;
             x = -z;
             y = t * (y2 - y1) + y1;
         }
         else
         if ((c & RIGHT) != 0)
         {
             t = (z1 - x1) / ((x2 - x1) - (z2 - z1));
             z = t * (z2 - z1) + z1;
             x = z;
             y = t * (y2 - y1) + y1;
         }
         else
         if ((c & BOTTOM) != 0)
         {
             t = (z1 + y1) / ((y1 - y2) - (z2 - z1));
             z = t * (z2 - z1) + z1;
             x = t * (x2 - x1) + x1;
             y = -z;
         }
         else
         if ((c & TOP) != 0)
         {
             t = (z1 - y1) / ((y2 - y1) - (z2 - z1));
             z = t * (z2 - z1) + z1;
             x = t * (x2 - x1) + x1;
             y = z;
         }

         if (c == c1)
         {
             x1 = x;
             y1 = y;
             z1 = z;
             c1 = code (x, y, z);
         }
         else
         {
             x2 = x;
             y2 = y;
             z2 = z;
             c2 = code (x, y, z);
         }
      }

      showLine (g, x1, y1, z1, x2, y2, z2);
   }

   private int code (double x, double y, double z)
   {
      int result = 0;

      if (x < -z)
          result = LEFT;

      if (x > z)
          result |= RIGHT;

      if (y < -z)
          result |= BOTTOM;

      if (y > z)
          result |= TOP;

      return result;
   }

   private void identity ()
   {
      for (int row = 0; row < 4; row++)
           for (int col = 0; col < 4; col++)
                transform.setValue (row, col, (row == col) ? 1.0 : 0.0);
   }

   private void lineto (Graphics g, double x, double y, double z)
   {
      double xc = x*transform.getValue (0, 0)+
                  y*transform.getValue (1, 0)+
                  z*transform.getValue (2, 0)+
                  transform.getValue (3, 0);

      double yc = x*transform.getValue (0, 1)+
                  y*transform.getValue (1, 1)+
                  z*transform.getValue (2, 1)+
                  transform.getValue (3, 1);

      double zc = x*transform.getValue (0, 2)+
                  y*transform.getValue (1, 2)+
                  z*transform.getValue (2, 2)+
                  transform.getValue (3, 2);

      clip (g, cx, cy, cz, xc, yc, zc);

      cx = xc;
      cy = yc;
      cz = zc;
   }

   private void moveto (double x, double y, double z)
   {
      cx = x*transform.getValue (0, 0)+
           y*transform.getValue (1, 0)+
           z*transform.getValue (2, 0)+
           transform.getValue (3, 0);

      cy = x*transform.getValue (0, 1)+
           y*transform.getValue (1, 1)+
           z*transform.getValue (2, 1)+
           transform.getValue (3, 1);

      cz = x*transform.getValue (0, 2)+
           y*transform.getValue (1, 2)+
           z*transform.getValue (2, 2)+
           transform.getValue (3, 2);
   }

   private void rotateX (double angle)
   {
      angle = Math.toRadians (angle);

      Matrix r = new Matrix (4, 4);
      r.setValue (0, 0, 1.0);
      r.setValue (1, 1, Math.cos (angle));
      r.setValue (2, 1, Math.sin (angle));
      r.setValue (1, 2, -Math.sin (angle));
      r.setValue (2, 2, Math.cos (angle));
      r.setValue (3, 3, 1.0);

      transform = transform.multiply (r);
   }

   private void rotateY (double angle)
   {
      angle = Math.toRadians (angle);

      Matrix r = new Matrix (4, 4);
      r.setValue (0, 0, Math.cos (angle));
      r.setValue (2, 0, -Math.sin (angle));
      r.setValue (1, 1, 1.0);
      r.setValue (0, 2, Math.sin (angle));
      r.setValue (2, 2, Math.cos (angle));
      r.setValue (3, 3, 1.0);

      transform = transform.multiply (r);
   }

   private void rotateZ (double angle)
   {
      angle = Math.toRadians (angle);

      Matrix r = new Matrix (4, 4);
      r.setValue (0, 0, Math.cos (angle));
      r.setValue (1, 0, Math.sin (angle));
      r.setValue (0, 1, -Math.sin (angle));
      r.setValue (1, 1, Math.cos (angle));
      r.setValue (2, 2, 1.0);
      r.setValue (3, 3, 1.0);

      transform = transform.multiply (r);
   }

   private void scale (double sx, double sy, double sz)
   {
      Matrix s = new Matrix (4, 4);
      s.setValue (0, 0, sx);
      s.setValue (1, 1, sy);
      s.setValue (2, 2, sz);
      s.setValue (3, 3, 1.0);

      transform = transform.multiply (s);
   }

   private void setViewport (int x1, int y1, int x2, int y2)
   {
      height = y2-y1+1;

      double vxl = x1;
      double vyb = y1;
      double vxr = x2;
      double vyt = y2;

      vcx = (vxl+vxr)/2;
      vcy = (vyb+vyt)/2;

      vsx = (vxr-vxl)/2;
      vsy = (vyt-vyb)/2;
   }

   private void showLine (Graphics g, double xc1, double yc1, double zc1,
                          double xc2, double yc2, double zc2)
   {
      int xs1 = (int) ((xc1/zc1)*vsx+vcx);
      int ys1 = (int) ((yc1/zc1)*vsy+vcy);

      int xs2 = (int) ((xc2/zc2)*vsx+vcx);
      int ys2 = (int) ((yc2/zc2)*vsy+vcy);

      g.drawLine (xs1, height-ys1, xs2, height-ys2);
   }

   private void translate (double tx, double ty, double tz)
   {
      Matrix t = new Matrix (4, 4);
      t.setValue (0, 0, 1.0);
      t.setValue (3, 0, tx);
      t.setValue (1, 1, 1.0);
      t.setValue (3, 1, ty);
      t.setValue (2, 2, 1.0);
      t.setValue (3, 2, tz);
      t.setValue (3, 3, 1.0);

      transform = transform.multiply (t);
   }

   private class Edge
   {
      String startVertex, endVertex;

      Edge (String startVertex, String endVertex)
      {
         this.startVertex = startVertex;
         this.endVertex = endVertex;
      }
   }

   private class Matrix
   {
      private double [][] matrix;

      Matrix (int nrows, int ncols)
      {
         matrix = new double [nrows][];
         for (int row = 0; row < ncols; row++)
              matrix [row] = new double [ncols];
      }

      int getCols ()
      {
          return matrix [0].length;
      }

      int getRows ()
      {
          return matrix.length;
      }

      double getValue (int row, int col)
      {
         return matrix [row] [col];
      }

      void setValue (int row, int col, double value)
      {
         matrix [row] [col] = value;
      }

      Matrix multiply (Matrix m)
      {
         Matrix result = new Matrix (matrix.length, matrix [0].length);

         for (int i = 0; i < getRows (); i++)
              for (int j = 0; j < m.getCols (); j++)
                   for (int k = 0; k < m.getRows (); k++)
                        result.setValue (i, j, result.getValue (i, j) +
                                         matrix [i][k] * m.getValue (k, j));

         return result;
      }
   }

   private class Vertex
   {
      double x, y, z;

      Vertex (double x, double y, double z)
      {
         this.x = x;
         this.y = y;
         this.z = z;
      }
   }
}
